package ac.essex.ooechs.facedetection.boosting;

import ac.essex.ooechs.adaboost.AdaBoost;
import ac.essex.ooechs.adaboost.AdaBoostSample;
import ac.essex.ooechs.adaboost.AdaBoostLearner;
import ac.essex.ooechs.adaboost.AdaBoostSolution;
import ac.essex.ooechs.facedetection.util.AbstractDetectionProblem;
import ac.essex.ooechs.facedetection.util.IntegralTrainingImage;
import ac.essex.ooechs.facedetection.util.DataSet;
import ac.essex.ooechs.facedetection.singlestage.OneStageDetectionProblem;
import ac.essex.ooechs.facedetection.singlestage.OneStageSingleFeatureProblem;
import ac.essex.ooechs.facedetection.evolved.gp.GPFeatureFinderBoolean;
import ac.essex.ooechs.imaging.commons.fast.IntegralImage;
import ac.essex.gp.Evolve;
import ac.essex.gp.problems.DataStack;
import ac.essex.gp.individuals.Individual;
import ac.essex.gp.interfaces.console.ConsoleListener;

import java.io.File;

/**
 * <p/>
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version,
 * provided that any use properly credits the author.
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details at http://www.gnu.org
 * </p>
 *
 * @author Olly Oechsle, University of Essex, Date: 26-Jun-2008
 * @version 1.0
 */
public class BoostedDetectionProblem extends AdaBoost {

    protected AbstractDetectionProblem problem;

    public static void main(String[] args) throws Exception {
        System.out.println("Viola Jones Detection Problem");
        DataSet training = new DataSet("/home/ooechs/Desktop/pipe-images2/foreground", "/home/ooechs/Desktop/pipe-images2/background");
        DataSet testing = null;
        AdaBoost a = new BoostedDetectionProblem(new OneStageSingleFeatureProblem(training, testing));
        a.boost(10);
        AdaBoostSolution s = a.getBestSolution();
        s.save(new File("/home/ooechs/Desktop/vj.solution"));
    }

    public BoostedDetectionProblem(AbstractDetectionProblem problem) {
        this.problem = problem;
        problem.generations = 50;
    }

    public AdaBoostSample[] getSamples() {
        AdaBoostSample[] samples = new AdaBoostSample[problem.trainingData.size()];
        for (int i = 0; i < problem.trainingData.size(); i++) {
            samples[i] =  problem.trainingData.elementAt(i);
        }
        return samples;
    }

    protected AdaBoostLearner weakLearn(AdaBoostSample[] adaBoostSamples, double[] weights) {
        // update the weight
        for (int i = 0; i < weights.length; i++) {
            problem.trainingData.elementAt(i).weight = weights[i];
        }

        // run the GP problem
        Evolve e = new Evolve(problem, new ConsoleListener(ConsoleListener.LOW_VERBOSITY));
        e.run();


        GPLearner l = new GPLearner(e.getBestIndividual());

        evaluate(l, e.getBestIndividual());

        // return the individual, wrapped as a GP learner.
        return l;

    }

    public void evaluate(GPLearner l, Individual ind) {

        DataStack dataStack = new DataStack();

        int totalTP = 0;
        int totalTN = 0;

        int TP = 0;
        int TN = 0;

        int mistakes = 0;

        for (int i = 0; i < problem.trainingData.size(); i++) {
            IntegralTrainingImage integralTrainingImage = problem.trainingData.elementAt(i);

            //dataStack.setIntegralImage(integralTrainingImage.getImage());*/

        AdaBoostSample[] samples = getSamples();

        //for (int i = 0; i < samples.length; i++) {
            AdaBoostSample sample = samples[i];
            dataStack.setIntegralImage((IntegralImage) sample.getData());

            if (sample.getLabel() != integralTrainingImage.getLabel()) {
                System.out.println("ARG");
            }



            double raw = ind.execute(dataStack);

            int result;

            if (ind.getPCM() != null) {
                result = ind.getPCM().getClassFromOutput(raw);
                if (result < 0) result = 0;
            } else {
                result = raw > 0 ? DataSet.TRUE : DataSet.FALSE;
            }

            if (result != l.classify(sample, null)) {
                System.out.println("ARG2");
            } 

            if (result == integralTrainingImage.classID) {
                if (integralTrainingImage.classID == DataSet.TRUE) TP++;
                if (integralTrainingImage.classID == DataSet.FALSE) TN++;
            } else {
                mistakes++;
            }

            if (integralTrainingImage.classID == DataSet.TRUE) {
                totalTP++;
            } else {
                totalTN++;
            }

        }

        System.out.print(TP / (double) totalTP);
        System.out.print("," + TN / (double) totalTN);
        System.out.print("," + (mistakes / (double) problem.trainingData.size()));
        System.out.println("\nTP: " + TP);

    }

}
