package ac.essex.ooechs.facedetection.experiments.eyes;

import ac.essex.ooechs.facedetection.util.DataSet;
import ac.essex.ooechs.facedetection.util.AbstractDetectionProblem;
import ac.essex.ooechs.facedetection.singlestage.OneStageDetectionProblem;
import ac.essex.ooechs.facedetection.singlestage.OneStageDetectionProblemBoolean;
import ac.essex.ooechs.facedetection.twostage.TwoStageEvolutionDRS;
import ac.essex.ooechs.facedetection.evolved.ga.results.evolved_pipes.*;
import ac.essex.gp.Evolve;
import ac.essex.gp.params.GPParams;
import ac.essex.gp.interfaces.console.ConsoleListener;

import java.io.File;

/**
 * Tests the difference between two approaches to learning the CBCL data set.
 * The first is using a one stage approach. The second uses pre-defined features
 * found using a GA.
 *
 * Both experiments are replicated 10 times with different random seeds, and
 * each GP run is 150 generations.
 *
 * Preliminary experiments show that DRS is the best technique.
 *
 * @author Olly Oechsle, University of Essex, Date: 20-Jun-2008
 * @version 1.0
 */
public class Experiment3_Eyes {

    public static void main(String[] args) {
        new Experiment3_Eyes().run();
    }

    public void run() {
        DataSet training = new DataSet("/home/ooechs/Data/me/smile/mouth/smiley/big", "/home/ooechs/Data/me/smile/notmouth");
        DataSet testing = null;

        int[] seeds = new int[1];
        for (int i = 0; i < seeds.length; i++) {
            seeds[i] = 2358+i;
        }

        System.out.println("ONE STAGE EVOLUTION");

        for (int i = 0; i < seeds.length; i++) {
             int seed = seeds[i];
             System.out.println("Running: " + i + ", seed=" + seed + " --------------------------");
             Evolve.seed =seed;
             OneStageDetectionProblem.generations = 25;
             AbstractDetectionProblem p = new OneStageDetectionProblemBoolean(training, testing);
             Evolve e = new Evolve(p, new ConsoleListener(ConsoleListener.LOW_VERBOSITY));
             e.run();
            e.getBestIndividual().save(new File("/home/ooechs/Desktop/saved.solution"));
             System.out.println(e.getBestIndividual().toJava());
            p.test(e.getBestIndividual());
         }

        System.out.println("Done :)");
    }

    class TwoStageEvolutionPipes extends TwoStageEvolutionDRS {

        public TwoStageEvolutionPipes(DataSet training, DataSet testing) {
            super(training, testing);
        }

        public void registerParams(GPParams params) {
            params.registerNode(new pipes_TwoRectFeature1());
            params.registerNode(new pipes_TwoRectFeature2());
            params.registerNode(new pipes_TwoRectFeature3());
            params.registerNode(new pipes_TwoRectFeature4());
            params.registerNode(new pipes_TwoRectFeature5());
            params.registerNode(new pipes_TwoRectFeature6());
            params.registerNode(new pipes_TwoRectFeature7());
            params.registerNode(new pipes_TwoRectFeature8());
            params.registerNode(new pipes_TwoRectFeature9());
            params.registerNode(new pipes_TwoRectFeature10());
            params.registerNode(new pipes_TwoRectFeature11());
            params.registerNode(new pipes_TwoRectFeature12());
            params.registerNode(new pipes_TwoRectFeature13());
            params.registerNode(new pipes_TwoRectFeature14());
            params.registerNode(new pipes_TwoRectFeature15());

            params.registerNode(new pipes_ThreeRectFeature1());
            params.registerNode(new pipes_ThreeRectFeature2());
            params.registerNode(new pipes_ThreeRectFeature3());
            params.registerNode(new pipes_ThreeRectFeature4());
            params.registerNode(new pipes_ThreeRectFeature5());
            params.registerNode(new pipes_ThreeRectFeature6());
            params.registerNode(new pipes_ThreeRectFeature7());
            params.registerNode(new pipes_ThreeRectFeature8());
            params.registerNode(new pipes_ThreeRectFeature9());
            params.registerNode(new pipes_ThreeRectFeature10());
            params.registerNode(new pipes_ThreeRectFeature11());
            params.registerNode(new pipes_ThreeRectFeature12());
            params.registerNode(new pipes_ThreeRectFeature13());
            params.registerNode(new pipes_ThreeRectFeature14());
            params.registerNode(new pipes_ThreeRectFeature15());

            params.registerNode(new pipes_FourRectFeature1());
            params.registerNode(new pipes_FourRectFeature2());
            params.registerNode(new pipes_FourRectFeature3());
            params.registerNode(new pipes_FourRectFeature4());
            params.registerNode(new pipes_FourRectFeature5());
            params.registerNode(new pipes_FourRectFeature6());
            params.registerNode(new pipes_FourRectFeature7());
            params.registerNode(new pipes_FourRectFeature8());
            params.registerNode(new pipes_FourRectFeature9());
            params.registerNode(new pipes_FourRectFeature10());
            params.registerNode(new pipes_FourRectFeature11());
            params.registerNode(new pipes_FourRectFeature12());
            params.registerNode(new pipes_FourRectFeature13());
            params.registerNode(new pipes_FourRectFeature14());
            params.registerNode(new pipes_FourRectFeature15());

/*        params.registerNode(new TwoRectVarFeature1());
        params.registerNode(new TwoRectVarFeature2());
        params.registerNode(new TwoRectVarFeature1());
        params.registerNode(new TwoRectVarFeature1());
        params.registerNode(new TwoRectVarFeature1());
        params.registerNode(new TwoRectVarFeature1());
        params.registerNode(new TwoRectVarFeature1());
        params.registerNode(new TwoRectVarFeature1());
        params.registerNode(new TwoRectVarFeature1());
        params.registerNode(new TwoRectVarFeature10());

        params.registerNode(new ThreeRectVarFeature1());
        params.registerNode(new ThreeRectVarFeature2());
        params.registerNode(new ThreeRectVarFeature3());
        params.registerNode(new ThreeRectVarFeature4());
        params.registerNode(new ThreeRectVarFeature5());
        params.registerNode(new ThreeRectVarFeature6());
        params.registerNode(new ThreeRectVarFeature7());
        params.registerNode(new ThreeRectVarFeature8());
        params.registerNode(new ThreeRectVarFeature9());
        params.registerNode(new ThreeRectVarFeature10());

        params.registerNode(new FourRectVarFeature1());
        params.registerNode(new FourRectVarFeature2());
        params.registerNode(new FourRectVarFeature3());
        params.registerNode(new FourRectVarFeature4());
        params.registerNode(new FourRectVarFeature5());
        params.registerNode(new FourRectVarFeature6());
        params.registerNode(new FourRectVarFeature7());
        params.registerNode(new FourRectVarFeature8());
        params.registerNode(new FourRectVarFeature9());
        params.registerNode(new FourRectVarFeature10());*/

        }

    }

}
