package ac.essex.ooechs.facedetection.singlestage;

import ac.essex.ooechs.facedetection.util.AbstractDetectionProblem;
import ac.essex.ooechs.facedetection.util.DataSet;
import ac.essex.ooechs.facedetection.util.IntegralTrainingImage;
import ac.essex.ooechs.facedetection.singlestage.nodes.*;
import ac.essex.gp.Evolve;
import ac.essex.gp.multiclass.PCM;
import ac.essex.gp.multiclass.BetterDRS;
import ac.essex.gp.multiclass.CachedOutput;
import ac.essex.gp.problems.DataStack;
import ac.essex.gp.individuals.Individual;
import ac.essex.gp.nodes.*;
import ac.essex.gp.nodes.ercs.SmallIntERC;
import ac.essex.gp.nodes.ercs.LargeIntERC;
import ac.essex.gp.nodes.ercs.PercentageERC;
import ac.essex.gp.nodes.ercs.CustomRangeParameterERC;
import ac.essex.gp.nodes.math.Pow;
import ac.essex.gp.params.GPParams;
import ac.essex.gp.params.NodeConstraints;
import ac.essex.gp.interfaces.console.ConsoleListener;

import java.util.Vector;

/**
 * <p/>
 * In this experiment we want to see how good GP is at solving
 * a face detection problem using a data set.
 * </p>
 *
 * @author Olly Oechsle, University of Essex, Date: 12-Jun-2008
 * @version 1.0
 */
public class OneStageSingleFeatureProblem extends AbstractDetectionProblem {

    protected int FPCounts[];
    protected int totalFP = 0;

    double bestFitness = Double.MAX_VALUE;

    static boolean weighting = false;
    static boolean optimisation = true;

    public static void main(String[] args) {

        DataSet training = new DataSet("/home/ooechs/Data/faces/train/face/", "/home/ooechs/Data/faces/train/non-face");
        DataSet testing = new DataSet("/home/ooechs/Data/faces/test/face/", "/home/ooechs/Data/faces/test/non-face");

        int[] seeds = new int[5];
        for (int i = 0; i < seeds.length; i++) {
            seeds[i] = 2357 + i;
        }

        /*      System.out.println("ERC OPTIMISATION OFF");
        optimisation = false;
        System.out.println("NO WEIGHTING --------------------------------");
        FaceDetectionProblem.weighting = false;

        for (int i = 0; i < seeds.length; i++) {
            int seed = seeds[i];
            System.out.println("Running: " + i + ", seed=" + seed);
            Evolve.seed =seed;
            FaceDetectionProblem p = new FaceDetectionProblem(training, testing);
            new Evolve(p, new ConsoleListener(ConsoleListener.SILENT)).run();
            System.gc();
        }

        System.out.println("WEIGHTING STARTED -------------------------");
        FaceDetectionProblem.weighting = true;

        for (int i = 0; i < seeds.length; i++) {
            int seed = seeds[i];
            System.out.println("Running: " + i + ", seed=" + seed);
            Evolve.seed =seed;
            FaceDetectionProblem p = new FaceDetectionProblem(training, testing);
            new Evolve(p, new ConsoleListener(ConsoleListener.SILENT)).run();
            System.gc();
        }

        System.out.println("ERC OPTIMISATION ON");
        optimisation = true;
        System.out.println("NO WEIGHTING --------------------------------");
        FaceDetectionProblem.weighting = false;

        for (int i = 0; i < seeds.length; i++) {
            int seed = seeds[i];
            System.out.println("Running: " + i + ", seed=" + seed);
            Evolve.seed =seed;
            FaceDetectionProblem p = new FaceDetectionProblem(training, testing);
            new Evolve(p, new ConsoleListener(ConsoleListener.SILENT)).run();
            System.gc();
        }*/

        System.out.println("WEIGHTING STARTED -------------------------");
        optimisation = true;
        weighting = false;

        for (int i = 0; i < seeds.length; i++) {
            int seed = seeds[i];
            System.out.println("Running: " + i + ", seed=" + seed);
            Evolve.seed = seed;
            OneStageSingleFeatureProblem p = new OneStageSingleFeatureProblem(training, testing);
            new Evolve(p, new ConsoleListener(ConsoleListener.LOW_VERBOSITY)).run();
            System.gc();
        }

        System.out.println("DONE ----");

/*
            FaceDetectionProblem.weighting = true;
            Evolve.seed =2361;
            FaceDetectionProblem p = new FaceDetectionProblem(training, testing);
            new Evolve(p).run();/*//*//**/

    }

    public OneStageSingleFeatureProblem(DataSet training, DataSet testing) {
        super(training, testing);
    }

    public String getName() {
        return "Face Detection Problem - Fixed Features";
    }

    public void initialise(Evolve evolve, GPParams params) {


        params.registerNode(new TwoHorizontalRectangleFeature());
        params.registerNode(new TwoVerticalRectangleFeature());

        params.registerNode(new ThreeHorizontalRectangleFeature());
        params.registerNode(new ThreeVerticalRectangleFeature());

        params.registerNode(new FourHorizontalRectangleFeature());
        params.registerNode(new FourVerticalRectangleFeature());

        params.registerNode(new TwoHorizontalRectangleVarianceFeature());
        params.registerNode(new TwoVerticalRectangleVarianceFeature());

        params.registerNode(new ThreeHorizontalRectangleVarianceFeature());
        params.registerNode(new ThreeVerticalRectangleVarianceFeature());

        params.registerNode(new FourHorizontalRectangleVarianceFeature());
        params.registerNode(new FourVerticalRectangleVarianceFeature());

        params.registerNode(new SmallIntERC());
        params.registerNode(new LargeIntERC());
        params.registerNode(new PercentageERC());

        params.registerNode(new CustomRangeParameterERC(1, 15));
        
        params.setReturnType(NodeConstraints.NUMBER);

    }

    public void customiseParameters(GPParams params) {
        params.setGenerations(generations);
        params.setERCOptimisationEnabled(optimisation);
    }


    public void onGenerationStart() {
        super.onGenerationStart();
        if (FPCounts != null) {
            // redistribute the weights
            for (int i = 0; i < trainingData.size(); i++) {
                IntegralTrainingImage trainingImage = trainingData.elementAt(i);
                // add the one to avoid divide by zero and ensure that everything has some weight
                int mistakeCount = FPCounts[i] + 1;

                // adapt the training image weight towards the new value
                double newValue = (mistakeCount * trainingData.size()) / (double) totalFP;

                double difference = newValue - trainingImage.weight;

                difference *= 0.33;

                // learning rate
                if (weighting) trainingImage.weight += difference;

                //trainingImage.weight = (mistakeCount * trainingData.size()) / (double) totalFP;
                //System.out.println("Training Image Weight: " + trainingImage.weight);
            }
        }
        FPCounts = new int[trainingData.size()];
        totalFP = 0;
    }

/*    public void setWeights() {

        double e = 0;
        boolean[] correct = new boolean[m];
        for (int i = 0; i < samples.length; i++) {
            correct[i] = h[t].classify(samples[i], null) == samples[i].getLabel();
            if (!correct[i]) e += d[i];
            // record the success of individual classifiers for reference
            else h[t].individualTP++;
        }

        // set the beta value associated with this hypothesis
        // using an exponential function that gives an exponential
        // distribution from 0 -> 1 for values 0 -> 0.5
        b[t] = e / (1 - e);

        // update the distribution. The lower the error, the lower b will be, so if instance i
        // was classified correctly it is multiplied by b, which reduces the weight of that sample.
        // If a sample is measured incorrectly, its weight is not changed.
        double dTotal = 0;
        for (int i = 0; i < samples.length; i++) {
            d[i] =  d[i] * (correct[i] ? b[t] : 1);
            dTotal += d[i];
        }

        // normalise the distribution so that all the weights add up to 1.0
        for (int i = 0; i < samples.length; i++) {
            d[i] /= dTotal;
        }

    }*/

    public void evaluate(Individual individual, DataStack dataStack, Evolve evolve) {

        // parameters for calculating fitness
        double totalTrueWeight = 0;
        double totalFalseWeight = 0;
        double trueWeight = 0;
        double falseWeight = 0;

        int totalTrue = 0;
        int totalFalse = 0;
        int TP = 0;
        int FP = 0;

        int hits = 0;
        int mistakes = 0;

        PCM pcm = new BetterDRS();

        dataStack.usesImaging = false;

        // go through all the training data
        for (int i = 0; i < trainingData.size(); i++) {

            IntegralTrainingImage trainingImage = trainingData.elementAt(i);

            dataStack.setIntegralImage(trainingImage.getImage());

            double result = individual.execute(dataStack);


            if (!dataStack.usesImaging) {
                // Important - do not include the mistakes of stupid individuals
                // otherwise the definition of difficulty and therefore the weights
                // of training data are inaccurate.
                individual.setWorstFitness();
                return;
            }

            pcm.addResult(result, trainingImage.classID);

        }

        // calculate the PCM thresholds
        pcm.calculateThresholds();

        // figure out the fitness
        Vector<CachedOutput> outputcache = pcm.getCachedResults();

        int lastResult = 0;
        boolean goodClassifier = false;
        for (int i = 0; i < outputcache.size(); i++) {
            CachedOutput cachedOutput = outputcache.elementAt(i);
            int result = pcm.getClassFromOutput(cachedOutput.rawOutput);
            if (result < 0) result = 0;
            if (lastResult != result) {
                goodClassifier = true;
                break;
            }
        }

        if (!goodClassifier) {
            individual.setWorstFitness();
            return;
        }

        for (int i = 0; i < outputcache.size(); i++) {

            CachedOutput cachedOutput = outputcache.elementAt(i);

            int result = pcm.getClassFromOutput(cachedOutput.rawOutput);

            IntegralTrainingImage trainingImage = trainingData.elementAt(i);

            boolean returnedTrue = result == 1;

            if (result == cachedOutput.expectedClass) {
                hits++;
            } else {
                mistakes++;
                FPCounts[i]++;
                totalFP++;
            }

            if (cachedOutput.expectedClass == 1) {
                totalTrueWeight += trainingImage.weight;
                totalTrue++;
                if (returnedTrue) {
                    trueWeight += trainingImage.weight;
                    TP++;
                }
            } else {
                totalFalseWeight += trainingImage.weight;
                totalFalse++;
                if (returnedTrue) {
                    falseWeight += trainingImage.weight;
                    FP++;
                }
            }

        }

        // make sure the individual doesn't hog too much memory
        pcm.clearCachedResults();

        double x = 1;

        double truePercentage = trueWeight / totalTrueWeight;
        double falsePercentage = falseWeight / totalFalseWeight;
        double weightedFitness = x * (1 - truePercentage) + falsePercentage;
        //double weightedFitness = falseWeight / totalFalseWeight;

/*
        double unweightedTruePercentage = TP / (double) totalTrue;
        double unweightedFalsePercentage = FP / (double) totalFalse;
        double unweightedFitness = x * (1 - unweightedTruePercentage) + unweightedFalsePercentage;
*/

        individual.setKozaFitness(weightedFitness);
        individual.setAlternativeFitness(mistakes);

 /*       if (unweightedFitness < bestFitness) {
            bestFitness = unweightedFitness;
            bestIndividual = individual;
        }*/

        individual.setHits(hits);
        individual.setMistakes(mistakes);
        individual.setPCM(pcm);

    }
}
