package ac.essex.ooechs.facedetection.twostage;

import ac.essex.gp.problems.DataStack;
import ac.essex.gp.Evolve;
import ac.essex.gp.individuals.Individual;
import ac.essex.gp.nodes.*;
import ac.essex.gp.nodes.logic.More;
import ac.essex.gp.nodes.logic.Less;
import ac.essex.gp.nodes.logic.Equals;
import ac.essex.gp.nodes.logic.Between;
import ac.essex.gp.nodes.ercs.*;
import ac.essex.gp.nodes.math.Pow;
import ac.essex.gp.params.GPParams;
import ac.essex.gp.params.NodeConstraints;
import ac.essex.gp.interfaces.console.ConsoleListener;
import ac.essex.ooechs.facedetection.util.DataSet;
import ac.essex.ooechs.facedetection.util.IntegralTrainingImage;
import ac.essex.ooechs.facedetection.evolved.ga.results.evolved_face.*;
import ac.essex.ooechs.facedetection.util.AbstractDetectionProblem;

/**
 * <p/>
 * In this experiment we want to see how good GP is at solving
 * a face detection problem using a data set.
 * </p>
 *
 * @author Olly Oechsle, University of Essex, Date: 12-Jun-2008
 * @version 1.0
 */
public class TwoStageEvolutionBoolean extends AbstractDetectionProblem {

    protected int FPCounts[];
    protected int totalFP = 0;

    double bestFitness = Double.MAX_VALUE;

    static boolean weighting = false;
    static boolean optimisation = true;

    public static void main(String[] args) {

        DataSet training = new DataSet("/home/ooechs/Data/faces/train/face/", "/home/ooechs/Data/faces/train/non-face");
        DataSet testing = new DataSet("/home/ooechs/Data/faces/test/face/", "/home/ooechs/Data/faces/test/non-face");

        int[] seeds = new int[1];
        for (int i = 0; i < seeds.length; i++) {
            seeds[i] = 2357+i;
        }

        //System.out.println("ERC OPTIMISATION IS OFF");

        optimisation = true;
        //weighting = true;

        for (int i = 0; i < seeds.length; i++) {
            int seed = seeds[i];
            System.out.println("Running: " + i + ", seed=" + seed + " --------------------------");
            Evolve.seed =seed;
            TwoStageEvolutionBoolean p = new TwoStageEvolutionBoolean(training, testing);
            new Evolve(p, new ConsoleListener(ConsoleListener.LOW_VERBOSITY)).run();
            System.gc();
        }
    }

    public TwoStageEvolutionBoolean(DataSet training, DataSet testing) {
        super(training, testing);
    }

    public String getName() {
        return "Two Stage Evolution using Pre-Evolved Features";
    }

    public void initialise(Evolve evolve, GPParams params) {


        params.registerNode(new If());
        params.registerNode(new BoolERC());
        params.registerNode(new More());
        params.registerNode(new Less());
        params.registerNode(new Equals());
        params.registerNode(new Between());

        params.registerNode(new Add());
        params.registerNode(new Sub());
        params.registerNode(new Mul());
        params.registerNode(new Div());

        params.registerNode(new Pow());
        params.registerNode(new Min());
        params.registerNode(new Max());

        params.registerNode(new TwoRectFeature1());
        params.registerNode(new TwoRectFeature2());
        params.registerNode(new TwoRectFeature3());
        params.registerNode(new TwoRectFeature4());
        params.registerNode(new TwoRectFeature5());
        params.registerNode(new TwoRectFeature6());
        params.registerNode(new TwoRectFeature7());
        params.registerNode(new TwoRectFeature8());
        params.registerNode(new TwoRectFeature9());
        params.registerNode(new TwoRectFeature10());

        params.registerNode(new ThreeRectFeature1());
        params.registerNode(new ThreeRectFeature2());
        params.registerNode(new ThreeRectFeature3());
        params.registerNode(new ThreeRectFeature4());
        params.registerNode(new ThreeRectFeature5());
        params.registerNode(new ThreeRectFeature6());
        params.registerNode(new ThreeRectFeature7());
        params.registerNode(new ThreeRectFeature8());
        params.registerNode(new ThreeRectFeature9());
        params.registerNode(new ThreeRectFeature10());

        params.registerNode(new FourRectFeature1());
        params.registerNode(new FourRectFeature2());
        params.registerNode(new FourRectFeature3());
        params.registerNode(new FourRectFeature4());
        params.registerNode(new FourRectFeature5());
        params.registerNode(new FourRectFeature6());
        params.registerNode(new FourRectFeature7());
        params.registerNode(new FourRectFeature8());
        params.registerNode(new FourRectFeature9());
        params.registerNode(new FourRectFeature10());

        params.registerNode(new SmallIntERC());
        params.registerNode(new LargeIntERC());
        params.registerNode(new PercentageERC());

        params.setReturnType(NodeConstraints.BOOLEAN);

    }

    public void customiseParameters(GPParams params) {
        params.setGenerations(generations);
        params.setERCOptimisationEnabled(optimisation);
    }

    public void onGenerationStart() {
        super.onGenerationStart();
        if (FPCounts != null) {
            // redistribute the weights
            for (int i = 0; i < trainingData.size(); i++) {
                IntegralTrainingImage trainingImage = trainingData.elementAt(i);
                // add the one to avoid divide by zero and ensure that everything has some weight
                int mistakeCount = FPCounts[i] + 1;

                // adapt the training image weight towards the new value
                double newValue = (mistakeCount * trainingData.size()) / (double) totalFP;

                double difference = newValue - trainingImage.weight;

                difference *= 0.33;

                // learning rate
                if (weighting) trainingImage.weight += difference;

                //trainingImage.weight = (mistakeCount * trainingData.size()) / (double) totalFP;
                //System.out.println("Training Image Weight: " + trainingImage.weight);
            }
        }
        FPCounts = new int[trainingData.size()];
        totalFP = 0;
    }

    public void evaluate(Individual individual, DataStack dataStack, Evolve evolve) {

        // parameters for calculating fitness
        double totalTrueWeight = 0;
        double totalFalseWeight = 0;
        double trueWeight = 0;
        double falseWeight = 0;

        int totalTrue = 0;
        int totalFalse = 0;
        int TP = 0;
        int FP = 0;

        int hits = 0;
        int mistakes = 0;

        dataStack.usesImaging = false;

        // go through all the training data
        for (int i = 0; i < trainingData.size(); i++) {

            IntegralTrainingImage trainingImage = trainingData.elementAt(i);

            dataStack.setIntegralImage(trainingImage.getImage());

            double result = individual.execute(dataStack);

            if (!dataStack.usesImaging ) {
                // Important - do not include the mistakes of stupid individuals
                // otherwise the definition of difficulty and therefore the weights
                // of training data are inaccurate.
                individual.setWorstFitness();
                return;
            }

            boolean returnedTrue = result == 1;

            if (result == trainingImage.classID) {
                hits++;
            } else {
                mistakes++;
                    FPCounts[i]++;
                    totalFP++;
            }

           if (trainingImage.classID == 1) {
                totalTrueWeight+=trainingImage.weight;
                totalTrue++;
                if (returnedTrue) {
                    trueWeight += trainingImage.weight;
                    TP++;
                }
            } else {
                totalFalseWeight+= trainingImage.weight;
                totalFalse++;
                if (returnedTrue) {
                    falseWeight += trainingImage.weight;
                    FP++;
                }
            }


        }

        double x = 1;

        double truePercentage = trueWeight / totalTrueWeight;
        double falsePercentage = falseWeight / totalFalseWeight;
        double weightedFitness = x * (1 - truePercentage) + falsePercentage;
        //double weightedFitness = falseWeight / totalFalseWeight;

/*
        double unweightedTruePercentage = TP / (double) totalTrue;
        double unweightedFalsePercentage = FP / (double) totalFalse;
        double unweightedFitness = x * (1 - unweightedTruePercentage) + unweightedFalsePercentage;
*/

        individual.setKozaFitness(weightedFitness);
        individual.setAlternativeFitness(mistakes);

 /*       if (individual.getKozaFitness() < bestFitness) {
            bestFitness = individual.getKozaFitness();
            bestIndividual = individual;
        }*/

        individual.setHits(hits);
        individual.setMistakes(mistakes);

    }


}
